import argparse
import glob
import os
import sys
from os import path
from typing import Dict, List, Optional, cast

import colorama
from termcolor_util import red, green, yellow, cyan, eprint

from version_manager.command_current_version import print_current_tag_version
from version_manager.command_version_list import (
    print_single_tracked_version,
    print_all_tracked_versions,
)
from version_manager.matchers.pattern import Pattern
from version_manager.options_set import get_parameter_values, get_parameters_from_file
from version_manager.settings_reader import read_settings_file


class ProgramArguments(object):
    display: List[str]
    tag_name: bool
    set_tag: str
    load: Optional[str]
    version: bool
    ignore_missing_parents: bool
    set: Optional[List[str]]
    all: bool


def main() -> None:
    colorama.init()

    parser = argparse.ArgumentParser(description="Versions processor")

    parser.add_argument(
        "--display",
        "-d",
        metavar="NAME",
        nargs=1,
        help="Display the version of a single tracked version.",
    )
    parser.add_argument(
        "--all",
        "-a",
        "--list",
        action="store_true",
        help="Display all the tracked versions and their values.",
    )
    parser.add_argument(
        "--set",
        "-s",
        nargs="+",
        metavar="NAME=VAL",
        help="Set values overriding what's in the yml files.",
    )
    parser.add_argument(
        "--load",
        "-l",
        metavar="FILE",
        help="Override versions from the given yml file.",
    )
    parser.add_argument(
        "-t",
        "--tag-name",
        "--tag",
        action="store_true",
        help="Get the current name to use in general tags. If the "
        "branch name can't be detected from the git repo, the "
        "$BRANCH_NAME environment variable will be used.",
    )
    parser.add_argument(
        "--ignore-missing-parents",
        action="store_true",
        help="Ignore missing parents, and simply don't patch the "
        "values. Upstream values are still being patched if existing.",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Show the program version (2021.6.1)",
    )
    parser.add_argument(
        "-st",
        "--set-tag",
        metavar="TAG",
        help="Override the current tag returned by nested calls of `vm -t`. "
             "This works by setting the VERSION_MANAGER_TAG in the environment "
             "of the current process.",
    )

    argv: ProgramArguments = cast(ProgramArguments, parser.parse_args(sys.argv[1:]))

    if argv.version:
        print(cyan("version-manager: 2021.6.1"))
        sys.exit(0)

    # update the current version name if needed
    if argv.set_tag:
        os.environ["VERSION_MANAGER_TAG"] = argv.set_tag

    if argv.tag_name:
        print_current_tag_version()
        sys.exit(0)

    default_settings_file = path.realpath(path.join(os.getcwd(), "versions.json"))
    override_parameters = get_parameters_from_file(argv.load)
    override_parameters = get_parameter_values(override_parameters, argv.set)
    versions_to_process = read_settings_file(
        default_settings_file,
        override_parameters,
        argv.ignore_missing_parents,
        cwd=os.path.dirname(default_settings_file),
        display_item=argv.display[0] if argv.display else None
    )

    # Display a single tracked version
    if argv.display:
        print_single_tracked_version(argv.display[0], versions_to_process)
        sys.exit(0)

    # Display all tracked versions.
    if argv.all:
        print_all_tracked_versions(versions_to_process)
        sys.exit(0)

    eprint(cyan("Running on %s" % sys.version))

    files_to_process: Dict[str, List[Pattern]] = dict()

    for tracked_version in versions_to_process:
        for file_name, version_pattern in tracked_version.files.items():
            resolved_names = glob.glob(file_name)

            if not resolved_names:
                print(red("Unable to find any files for glob %s." % file_name))
                sys.exit(2)

            for resolved_name in resolved_names:
                if resolved_name in files_to_process:
                    file_patterns = files_to_process[resolved_name]
                else:
                    file_patterns = []
                    files_to_process[resolved_name] = file_patterns

                file_patterns.append(version_pattern)

    for resolved_name, version_patterns in files_to_process.items():
        with open(resolved_name, "r", encoding="utf-8") as resolved_file:
            content = resolved_file.read()
            new_content = content

        print(cyan("Patching %s:" % resolved_name))

        for version_pattern in version_patterns:
            tracked_version = version_pattern.tracked_version
            print(green("* %s@%s" % (tracked_version.name, tracked_version.version)))

            new_content = version_pattern.apply_pattern(new_content)

            if version_pattern.match_count != version_pattern.expected_count:
                print(
                    red(
                        "Got %d matches instead of %d."
                        % (version_pattern.match_count, version_pattern.expected_count)
                    )
                )
                sys.exit(3)

        if content == new_content:
            print(
                cyan("Content for %s is not changed. Won't patch it." % resolved_name)
            )
            continue

        with open(resolved_name, "w", encoding="utf-8") as output:
            output.write(new_content)

        print(yellow("Updated %s" % resolved_name))

    colorama.deinit()
    sys.exit(0)


if __name__ == "__main__":
    main()



import re

from .pattern import Pattern, TrackedVersion
from .regex_pattern import RegExPattern


class MavenPattern(Pattern):
    RE = re.compile(r"^maven\:(.*?)\:(.*?)$")

    def __init__(self, tracked_version: TrackedVersion, expression: str) -> None:
        m = MavenPattern.RE.match(expression)

        if not m:
            raise Exception("Unable to parse %s as a maven pattern.", expression)

        regexp_value = (
            "(<groupId>%s</groupId>\\s*"
            "<artifactId>%s</artifactId>\\s*"
            "<version>)(.*?)(</version>)"
            % (re.escape(m.group(1)), re.escape(m.group(2)))
        )

        self.tracked_version = tracked_version
        self.regex_pattern = RegExPattern(tracked_version, regexp_value)

    def apply_pattern(self, input: str) -> str:
        return self.regex_pattern.apply_pattern(input)

    @property
    def match_count(self) -> int:
        return self.regex_pattern.match_count

    @property
    def expected_count(self) -> int:
        return self.regex_pattern.expected_count



import subprocess
import re
import os
from termcolor_util import eprint, red

DIVERGED_FROM_RELEASE = re.compile(r"^.+?-\d+-\S+$")


def print_current_tag_version() -> None:
    version_name = get_current_tag_version()
    print(version_name)


def get_current_tag_version() -> str:
    # If we have the VERSION_MANAGER_TAG in the environment,
    # we use it instead of whatever git tag, or derived name
    # from the branch.
    if "VERSION_MANAGER_TAG" in os.environ:
        return os.environ["VERSION_MANAGER_TAG"]

    # if we have BRANCH_NAME in the environment, we use that one,
    # since it's already found for us.
    if "BRANCH_NAME" in os.environ:
        env_branch_name = os.environ["BRANCH_NAME"]

        if "/" not in env_branch_name:
            return env_branch_name

        return f"0.0.0-{escape_tag_name(env_branch_name)}"

    # We try to find if we have an annotated git tag
    try:
        current_release_version: str = subprocess.check_output(
            ["git", "describe"]
        ).decode("utf-8").strip()

        if not DIVERGED_FROM_RELEASE.match(current_release_version):
            return current_release_version  # => we're on a tagged release
    except Exception as e:
        eprint(red(str(e)))

    # We try to find the current branch name
    current_branch_name: str = subprocess.check_output(
        ["git", "rev-parse", "--abbrev-ref", "HEAD"]
    ).decode("utf-8").strip()

    return f"0.0.0-{escape_tag_name(current_branch_name)}"


def escape_tag_name(name: str) -> str:
    return re.sub(r"[^A-Za-z-0-9]", "-", name)



import sys
from os import path
from typing import Dict, Optional

import yaml
from termcolor_util import red

from .matcher_builder import matcher_builder
from .matchers.pattern import TrackedVersionSet, TrackedVersion


def read_settings_file(
    settings_file: str,
    override_settings: Dict[str, str],
    ignore_missing_parents: bool,
    cwd: str,
    display_item: Optional[str] = None,
) -> TrackedVersionSet:
    """
    Read the configured versions from the files. If a version is defined in the
    override_settings, then that value is going to be used, instead of what's
    read from the file.

    This allows overwriting versions, regardless of where they're read from.
    """
    if not path.exists(settings_file):
        settings_file = path.join(path.dirname(settings_file), "versions.yml")

        if not path.exists(settings_file):
            report_missing_settings_file(settings_file)
            sys.exit(1)

    with open(settings_file, "r", encoding="utf-8") as stream:
        settings = list(yaml.safe_load_all(stream))[0]

    result = list()

    for name, tracked_entry in settings.items():
        if display_item and name != display_item:
            continue

        try:
            tracked_version: TrackedVersion = TrackedVersion(name)
            tracked_version.version = (
                override_settings[name]
                if name in override_settings
                else parse_version(
                    tracked_entry["version"],
                    override_settings,
                    ignore_missing_parents,
                    cwd=cwd,
                )
            )

            tracked_files = tracked_entry["files"] if "files" in tracked_entry else {}

            for file_name in tracked_files.keys():
                tracked_file = matcher_builder(
                    tracked_version, tracked_files[file_name]
                )
                tracked_version.files[file_name] = tracked_file
        except ParentNotFound as e:
            if ignore_missing_parents:
                continue

            raise Exception("Unable to find parent", e)
        except Exception as e:
            raise Exception("Unable to read value: %s" % name, e)
        else:
            result.append(tracked_version)

    return result


def report_missing_settings_file(settings_file: str) -> None:
    print(red("%s configuration file is missing." % settings_file))


# This import is intentionally at the end, because it's a cyclic import
from .parse_version import parse_version, ParentNotFound  # NOQA



import unittest

from version_manager.matchers.maven_pattern import MavenPattern
from version_manager.matchers.pattern import TrackedVersion


class TestMavenPattern(unittest.TestCase):
    def test_maven_pattern(self):
        maven_code = """
            <dependency>
                <groupId>com.ciplogic</groupId>
                <artifactId>test</artifactId>
                <version>1.0.0-SNAPSHOT</version>
            </dependency>
            <dependency>
                <groupId>com.ciplogic</groupId>
                <artifactId>test2</artifactId>
                <version>1.0.0-SNAPSHOT</version>
            </dependency>
        """

        tracked_version = TrackedVersion("test_version")
        tracked_version.version = "2.0"

        m = MavenPattern(tracked_version, "maven:com.ciplogic:test")
        result = m.apply_pattern(maven_code)

        expected_code = """
            <dependency>
                <groupId>com.ciplogic</groupId>
                <artifactId>test</artifactId>
                <version>2.0</version>
            </dependency>
            <dependency>
                <groupId>com.ciplogic</groupId>
                <artifactId>test2</artifactId>
                <version>1.0.0-SNAPSHOT</version>
            </dependency>
        """

        self.assertEqual(expected_code, result)
        self.assertEqual(1, m.expected_count)
        self.assertEqual(1, m.match_count)


if __name__ == "__main__":
    unittest.main()



from typing import Any

from .matchers.pattern import TrackedVersion, Pattern
from .matchers.array_pattern import ArrayPattern
from .matchers.regex_pattern import RegExPattern
from .matchers.match_counter import MatchCounter
from .matchers.maven_pattern import MavenPattern
from .matchers.string_pattern import StringPattern


def matcher_builder(tracked_version: TrackedVersion, file_item: Any) -> Pattern:
    if isinstance(file_item, list):
        file_items = map(lambda it: matcher_builder(tracked_version, it), file_item)

        return ArrayPattern(tracked_version, list(file_items))

    if "count" in file_item:
        expression = (
            file_item["match"] if "match" in file_item else file_item["expression"]
        )

        return MatchCounter(
            tracked_version,
            matcher_builder(tracked_version, expression),
            int(file_item["count"]),
        )

    if MavenPattern.RE.match(file_item):
        return MavenPattern(tracked_version, file_item)

    if StringPattern.RE.match(file_item):
        return StringPattern(tracked_version, file_item)

    return RegExPattern(tracked_version, file_item)



from termcolor_util import cyan, red, eprint
import sys

from version_manager.util_find import find
from version_manager.matchers.pattern import TrackedVersionSet


def print_single_tracked_version(
    version_name: str, versions_to_process: TrackedVersionSet
) -> None:
    tracked_version = find(lambda it: it.name == version_name, versions_to_process)

    if not tracked_version:
        eprint(
            red(
                "Tracked version '%s' does not exist. Available are: "
                "%s."
                % (
                    version_name,
                    ", ".join(map(lambda it: it.name, versions_to_process)),
                )
            )
        )
        sys.exit(1)

    print(tracked_version.version)


def print_all_tracked_versions(versions_to_process: TrackedVersionSet) -> None:
    for it in versions_to_process:
        print(cyan(it.name, bold=True), "=>", cyan(it.version))



import os
import re
import subprocess
from os import path
from typing import Dict

from version_manager.matchers.pattern import TrackedVersionSet
from version_manager.settings_reader import read_settings_file
from version_manager.util_find import find

setting_files: Dict[str, TrackedVersionSet] = dict()

PARENT_RE = re.compile(r"^parent:(.+)@(.+?)$")


class ParentNotFound(Exception):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


def parse_parent_path(
    version: str,
    cwd: str,
    overriden_settings: Dict[str, str],
    ignore_missing_parents: bool,
) -> str:
    items = PARENT_RE.match(version)

    if not items:
        raise Exception(
            "The version must be in the 'parent:path@propertyname' "
            "format, got instead: '%s'." % version
        )

    parent_versions_file_path = items.group(1)
    property_name = items.group(2)

    full_path = path.realpath(path.join(cwd, parent_versions_file_path))

    if not path.exists(full_path):
        raise ParentNotFound("Unable to find referenced file: %s" % full_path)

    if path.isdir(full_path):
        full_path = path.join(full_path, "versions.json")

    if full_path not in setting_files:
        setting_files[full_path] = read_settings_file(
            full_path,
            overriden_settings,
            ignore_missing_parents,
            cwd=os.path.dirname(full_path)
        )

    property_value = find(lambda it: it.name == property_name, setting_files[full_path])

    if not property_value:
        available_properties = ", ".join(
            map(lambda it: "%s@%s" % (it.name, it.version), setting_files[full_path])
        )

        raise Exception(
            "Property '%s' is not defined in %s settings file. "
            "Available properties are: %s"
            % (property_name, full_path, available_properties)
        )

    return property_value.version


def custom_rstrip(s: str) -> str:
    """
    A custom rstrip that only strips the last Enter if it exists,
    keeping the other whitespace characters. This allows outputing
    values with whitespace formatting at the end, while handling
    the most common case of a single echo with an enter.

    """
    if s and s[-1] == "\n":
        s = s[0:-1]

    return s


def parse_version_with_path(
    version: str,
    cwd: str,
    overriden_settings: Dict[str, str],
    ignore_missing_parents: bool,
) -> str:
    if not isinstance(version, str):
        raise Exception(
            "Got version a %s of type %s, in %s" % (version, type(version), cwd)
        )

    if version.startswith("parent:"):
        return parse_parent_path(
            version, cwd, overriden_settings, ignore_missing_parents
        )

    if "`" not in version and "$" not in version:
        return custom_rstrip(version)

    command = extract_command(version)
    result: str = subprocess.check_output(
            ["/bin/sh", "-c", command],
            env=os.environ,
            encoding="utf-8",
            cwd=cwd,
    )

    return custom_rstrip(result)


def parse_version(
    version: str,
    overriden_settings: Dict[str, str],
    ignore_missing_parents: bool,
    cwd: str,
) -> str:
    return parse_version_with_path(
        version, cwd, overriden_settings, ignore_missing_parents
    )


def extract_command(version: str) -> str:
    if version.startswith("`"):
        return version[1:-1]

    if version.startswith("$"):
        return version[2:-1]

    raise Exception(f"Wrong version sent as command: {version}")



import re

from .pattern import Pattern, TrackedVersion
from .regex_pattern import RegExPattern
from termcolor_util import eprint, yellow


class StringPattern(Pattern):
    RE = re.compile(r"^(.*?)(\^\^|##|\*\*)VERSION(##|\*\*|\$\$)(.*?)$")

    def __init__(self, tracked_version: TrackedVersion, expression: str) -> None:
        m = StringPattern.RE.match(expression)

        if not m:
            raise Exception("Unable to parse %s as a string pattern" % expression)

        if m.group(2) == "##" or m.group(3) == "#":
            eprint(
                yellow(
                    "Version matched using expression '%s' "
                    "still uses the old '##' notation for delimiting the "
                    "version. This is not supported anymore since # denotes "
                    "a comment in YAML. Use '**' instead." % expression
                )
            )

        regexp_value = (
            ("^()" if m.group(2) == "^^" else "(%s)" % re.escape(m.group(1)))
            + "(.*?)"
            + ("$" if m.group(3) == "$$" else "(%s)" % re.escape(m.group(4)))
        )

        self.tracked_version = tracked_version
        self.regex_pattern = RegExPattern(
            tracked_version, regexp_value, extra_flags=re.M
        )

    def apply_pattern(self, input_str: str) -> str:
        return self.regex_pattern.apply_pattern(input_str)

    @property
    def match_count(self) -> int:
        return self.regex_pattern.match_count

    @property
    def expected_count(self) -> int:
        return self.regex_pattern.expected_count



import re

from .pattern import Pattern, TrackedVersion


class RegExPattern(Pattern):
    def __init__(
        self,
        tracked_version: TrackedVersion,
        expression: str,
        extra_flags: int = re.M | re.S,
    ) -> None:
        try:
            self.RE = re.compile(expression, extra_flags)
        except Exception as e:
            raise Exception("Unable to compile regex expression: %s" % expression, e)
        self.tracked_version = tracked_version
        self._match_count = 0

    def apply_pattern(self, text: str) -> str:
        """
        Apply the regex pattern on the text, replacing the 2nd
        group from every match.
        """
        found_matches = []
        last_index = -1

        m = self.RE.search(text, pos=last_index)
        while m:
            self._match_count += 1
            last_index = m.start() + len(m.group(0))

            found_matches.append(m)

            m = self.RE.search(text, pos=last_index)

        original_index: int = 0
        original_text: str = text

        result: str = ""

        for match in found_matches:
            result += (
                original_text[original_index : match.start()]
                + match.group(1)
                + self.tracked_version.version
                + (match.group(3) if len(match.groups()) >= 3 else "")
            )

            original_index = match.start() + len(match.group(0))

        result += original_text[original_index : len(original_text)]

        return result

    @property
    def match_count(self) -> int:
        return self._match_count

    @property
    def expected_count(self) -> int:
        return 1



from typing import Dict, List
from abc import ABCMeta, abstractmethod, abstractproperty


class Pattern(metaclass=ABCMeta):
    tracked_version: "TrackedVersion"

    @abstractmethod
    def apply_pattern(self, input: str) -> str:
        pass

    @abstractproperty
    def match_count(self) -> int:
        pass

    @abstractproperty
    def expected_count(self) -> int:
        pass


class TrackedVersion:
    name: str
    version: str
    files: Dict[str, Pattern]

    def __init__(self, name: str) -> None:
        self.name = name
        self.version = ""
        self.files = dict()


TrackedVersionSet = List[TrackedVersion]



from typing import List

from .pattern import Pattern, TrackedVersion


class ArrayPattern(Pattern):
    def __init__(
        self, tracked_version: TrackedVersion, delegate_patterns: List[Pattern]
    ) -> None:
        self.tracked_version = tracked_version
        self.delegate_patterns = delegate_patterns

    def apply_pattern(self, input: str) -> str:
        for pattern in self.delegate_patterns:
            input = pattern.apply_pattern(input)

        return input

    @property
    def match_count(self) -> int:
        return sum(map(lambda it: it.match_count, self.delegate_patterns))

    @property
    def expected_count(self) -> int:
        return sum(map(lambda it: it.expected_count, self.delegate_patterns))



from typing import Dict, List, Optional
import yaml


def get_parameter_values(
    parameter_values: Dict[str, str], values_list: Optional[List[str]]
) -> Dict[str, str]:
    """
    Override the parameter values that are given in the list.
    It assumes each parameter is in the 'KEY=VALUE' format.
    """
    if not values_list:
        return parameter_values

    for value in values_list:
        tokens = value.split("=", 2)
        parameter_values[tokens[0]] = tokens[1]

    return parameter_values


def get_parameters_from_file(file_name: Optional[str]) -> Dict[str, str]:
    if not file_name:
        return dict()

    with open(file_name, "r", encoding="utf-8") as stream:
        result = list(yaml.safe_load_all(stream))[0]

    return result



import unittest
from textwrap import dedent

from version_manager.matchers.regex_pattern import RegExPattern
from version_manager.matchers.pattern import TrackedVersion


class TestRegExPattern(unittest.TestCase):
    def test_double_replace(self):
        tracked_version = TrackedVersion("test_version")
        tracked_version.version = "zz"

        r = RegExPattern(tracked_version, "(aa)(bb)")
        result = r.apply_pattern("aabb")

        self.assertEqual("aazz", result)
        self.assertEqual(1, r.match_count)
        self.assertEqual(1, r.expected_count)

    def test_triple_replace_multiple_matches(self):
        """
        Just a normal replace on the version that uses all
        the three groups.
        """
        tracked_version = TrackedVersion("test_version")
        tracked_version.version = "zz"
        tracked_version.expected_count = 2

        r = RegExPattern(tracked_version, "^(.*?)(bb)(.*?)$")

        result = r.apply_pattern(
            dedent(
                """
            aa bb cc
            this is a line
            what bb is this
        """
            )
        )

        self.assertEqual(
            dedent(
                """
            aa zz cc
            this is a line
            what zz is this
        """
            ),
            result,
        )
        self.assertEqual(2, r.match_count)
        self.assertEqual(1, r.expected_count)


if __name__ == "__main__":
    unittest.main()



from typing import Callable, Iterable, TypeVar, Union


T = TypeVar("T")
N = Union[T, None]


def find(check: Callable[[T], bool], items: Iterable[T]) -> N:
    for item in items:
        if check(item):
            return item

    return None



from setuptools import setup
from setuptools import find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

packages = find_packages()

setup(
    name="vm",
    version="2021.6.1",
    description="version_manager",
    long_description=readme,
    author="Bogdan Mustiata",
    author_email="bogdan.mustiata@gmail.com",
    license="BSD",
    entry_points={
        "console_scripts": [
            "version-manager=version_manager.mainapp:main",
            "vm=version_manager.mainapp:main",
        ]
    },
    install_requires=[
        "colorama == 0.4.3",
        "termcolor_util == 1.2.0",
        "PyYAML == 5.1.2",
    ],
    packages=packages,
    package_data={"": ["*.txt", "*.rst"]},
)



from .pattern import TrackedVersion, Pattern


class MatchCounter(Pattern):
    def __init__(
        self,
        tracked_version: TrackedVersion,
        delegate_pattern: Pattern,
        expected_count: int,
    ) -> None:
        self.tracked_version = tracked_version
        self.delegate_pattern = delegate_pattern
        self._expected_count = expected_count

    def apply_pattern(self, input: str) -> str:
        return self.delegate_pattern.apply_pattern(input)

    @property
    def match_count(self) -> int:
        if self._expected_count < 0:
            return self._expected_count

        return self.delegate_pattern.match_count

    @property
    def expected_count(self) -> int:
        return self._expected_count
